/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Clippings extension for Mozilla Firefox.
 *
 * The Initial Developer of the Original Code is 
 * Alex Eng <ateng@users.sourceforge.net>.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** */

var gStatusBar;
var gCurrentListItemIndex = -1;

// XPCOM RDF services
var gRDFSvc = gClippings.RDFSvc;
var gRDFContainerUtils = gClippings.RDFContainerUtils;

// Undo
var gUndoStack = {
  length: 0,
  _stack: [],

  push: function (aAction, aURI, aName, aText)
  {
    this._stack.push({action: aAction, uri: aURI, name: aName, text: aText});
    this.length++;
  },

  pop: function ()
  {
    var rv = this._stack.pop();
    this.length--;
    return rv;
  }
};

// Redo - only 1 undo action is reversible
var gRedoStack = {
  length:   0,
  _lastUndo: null,

  push: function (aAction, aURI, aName, aText)
  {
    this._lastUndo = {action: aAction, uri: aURI, name: aName, text: aText};
    this.length = (this.length == 0 ? 1 : 1);
  },

  pop: function ()
  {
    var rv = {
      action: this._lastUndo.action,
      uri:    this._lastUndo.uri,
      name:   this._lastUndo.name,
      text:   this._lastUndo.text
    };
    this._lastUndo = null;
    this.length = 0;
    return rv;
  }
};


// Flags for gUndoStack._stack[i].action
const ACTION_EDITNAME = 1;
const ACTION_EDITTEXT = 2;
const ACTION_DELETE = 3;
const ACTION_CREATENEW = 4;

// Flags for aDestUndoStack parameter of functions for reversible actions
const UNDO_STACK = 1;
const REDO_STACK = 2;


function init() 
{
  var clippingsList = document.getElementById("clippings-list");
  var rootNode = gRDFSvc.GetResource("http://clippings.mozdev.org/rdf/user-clippings");

  try {
    // If Clippings data file doesn't exist, it will be created automatically.
    gClippings.dataSrc = gRDFSvc.GetDataSourceBlocking(gClippings.getCurrentProfileURL() + "/" + CLIPDAT_FILE_NAME);
  }
  catch (e) {
    alert(e);
  }
  clippingsList.database.AddDataSource(gClippings.dataSrc);

  try {
    // MakeSeq will create a new <RDF:Seq> if it doesn't already exist.
    // If it does, it just initializes it (similar to calling gClippings.RDFContainer.Init()).
    gClippings.RDFContainer = gRDFContainerUtils.MakeSeq(gClippings.dataSrc, rootNode);
    gClippings.RDFContainer = gClippings.RDFContainer.QueryInterface(Components.interfaces.nsIRDFContainer);
  }
  catch (e) {
    alert(e);
  }

  var numEntries = gClippings.RDFContainer.GetCount();
  document.getElementById("entry-properties").selectedIndex = numEntries == 0 ? 1 : 2;
    
  clippingsList.builder.rebuild();
  gStatusBar = document.getElementById("app-status");
  gStatusBar.label = numEntries + " item(s)";

  if (numEntries > 0) {
    clippingsList.selectedIndex = 0;
    gCurrentListItemIndex = 0;
    clippingsList.click();
  }
}


function newClipping() 
{
  updateCurrentClippingData();
  newClippingHelper("New Entry", "", null, UNDO_STACK);
}


function newClippingHelper(aName, aText, aURI, aDestUndoStack)
{
  var newNodeURI;
  if (aURI) {
    newNodeURI = gClippings.createNewClippingWithURI(aURI, aName, aText);
  }
  else {
    newNodeURI = gClippings.createNewClipping(aName, aText);
  }

  var newListItem;
  while (! (newListItem = document.getElementById(newNodeURI))) { 
    // Delay selection of the new entry until its corresponding <listitem>
    // element is created.
  }
  var clippingsList = document.getElementById("clippings-list");
  clippingsList.selectItem(newListItem);
  clippingsList.ensureElementIsVisible(newListItem);
  clippingsList.click();

  var deck = document.getElementById("entry-properties");
  if (deck.selectedIndex != 0) {
    deck.selectedIndex = 0;
  }

  var clippingName = document.getElementById("clipping-name");
  clippingName.select();
  clippingName.focus();

  if (aDestUndoStack == UNDO_STACK) {
    gUndoStack.push(ACTION_CREATENEW, newNodeURI, aName, aText);
    gClippings.log(String.format("New entry %S added to undo stack", aName));
  }
  else if (aDestUndoStack == REDO_STACK) {
    gRedoStack.push(ACTION_CREATENEW, newNodeURI, aName, aText);
    gClippings.log(String.format("New entry %S added to redo stack", aName));
  }
}


function pasteClippingAsNew()
{
  updateCurrentClippingData();

  var newNodeURI = gClippings.newFromClipboard();
  if (! newNodeURI) {
    gClippings.beep();
    gStatusBar.label = "Clipboard is empty or does not contain textual data.";
    return;
  }

  var newListItem;
  while (! (newListItem = document.getElementById(newNodeURI))) { 
    // Delay selection of the new entry until its corresponding <listitem>
    // element is created.
  }
  var clippingsList = document.getElementById("clippings-list");
  clippingsList.selectItem(newListItem);
  clippingsList.ensureElementIsVisible(newListItem);
  clippingsList.click();

  var deck = document.getElementById("entry-properties");
  if (deck.selectedIndex != 0) {
    deck.selectedIndex = 0;
  }

  var clippingName = document.getElementById("clipping-name");
  clippingName.select();
  clippingName.focus();

  gUndoStack.push(ACTION_CREATENEW, newNodeURI, 
		  newListItem.getAttribute("clipName"), 
		  newListItem.getAttribute("clipText"));
  gClippings.log(String.format("New entry added to undo stack\nName = %S", 
			       newListItem.getAttribute("clipName")));
}


function updateEditStatus() 
{
  gStatusBar.label = "Edit entry";
}


function saveClipping() 
{
  updateCurrentClippingData();
  gStatusBar.label = "Saving...";

  try {
    gClippings.dataSrc.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
  }
  catch (e) {
    gStatusBar.label = "Error: Save Failed!";
    alert(e);
    return;
  }

  gStatusBar.label = "Saved";
}


function deleteClipping() 
{
  var clippingsList = document.getElementById("clippings-list");
  if (clippingsList.getRowCount() == 0) {
    gClippings.beep();
    gStatusBar.label = "Nothing to delete";
    return;
  }

  updateCurrentClippingData();

  var currentItem = clippingsList.selectedItem;
  if (! currentItem) {
    return;
  }

  deleteClippingHelper(currentItem, UNDO_STACK);
}


function deleteClippingHelper(aListItem, aDestUndoStack)
{
  var clippingsList = document.getElementById("clippings-list");
  var deletedItemIndex = clippingsList.selectedIndex;
  var entry = {
    uri: aListItem.id,
    name: aListItem.getAttribute("clipName"),
    text: aListItem.getAttribute("clipText")
  };
  var subjectNode = gRDFSvc.GetResource(entry.uri);
  var predName = gRDFSvc.GetResource("http://clippings.mozdev.org/ns/rdf#name");
  var targName = gRDFSvc.GetLiteral(entry.name);
  var predText = gRDFSvc.GetResource("http://clippings.mozdev.org/ns/rdf#text");
  var targText = gRDFSvc.GetLiteral(entry.text);

  try {
    clippingsList.clearSelection();
  }
  catch (e) {
    // Calling method clearSelection() returns an error, although it works anyway. 
  }

  if (aDestUndoStack == UNDO_STACK) {
    gUndoStack.push(ACTION_DELETE, entry.uri, entry.name, entry.text);
    gClippings.log(String.format("Deleted entry %S (URI %s added to undo stack)", entry.name, entry.uri));
  }
  else if (aDestUndoStack == REDO_STACK) {
    gRedoStack.push(ACTION_DELETE, entry.uri, entry.name, entry.text);
    gClippings.log(String.format("Deleted entry %S (URI %s added to redo stack)", entry.name, entry.uri));
  }

  gClippings.dataSrc.Unassert(subjectNode, predName, targName);
  gClippings.dataSrc.Unassert(subjectNode, predText, targText);
  gClippings.RDFContainer.RemoveElement(subjectNode, true);

  clippingsList.builder.rebuild();
  var numEntries = gClippings.RDFContainer.GetCount();
  gStatusBar.label = numEntries + " item(s)";

  var numItems = clippingsList.getRowCount();
  if (numItems == 0) {
    document.getElementById("entry-properties").selectedIndex = 1;
    gCurrentListItemIndex = -1;
  }
  else if (deletedItemIndex == numItems) {  // Deleted entry was last list item
    clippingsList.selectedIndex = numItems - 1;
    clippingsList.ensureIndexIsVisible(numItems - 1);
    updateDisplay();
  }
  else {
    clippingsList.selectedIndex = deletedItemIndex;
    clippingsList.ensureIndexIsVisible(deletedItemIndex);
    updateDisplay();
  }
}


function updateCurrentEntryStatus()
{
  var clippingsList = document.getElementById("clippings-list");
  if (clippingsList.getRowCount() == 0) {
    return;
  }

  var numEntries = gClippings.RDFContainer.GetCount();
  gStatusBar.label = numEntries + " item(s)";

  var deck = document.getElementById("entry-properties");
  if (deck.selectedIndex != 0) {
    deck.selectedIndex = 0;
  }
}


function updateDisplay()
{
  var clippingName = document.getElementById("clipping-name");
  var clippingText = document.getElementById("clipping-text");
  var clippingsList = document.getElementById("clippings-list");
  clippingName.value = clippingsList.selectedItem.getAttribute("clipName");
  clippingText.value = clippingsList.selectedItem.getAttribute("clipText");

  gCurrentListItemIndex = clippingsList.selectedIndex;
}


function updateName(aName)
{
  var clippingsList = document.getElementById("clippings-list");
  var currentItem = clippingsList.getItemAtIndex(gCurrentListItemIndex);
  if (! currentItem) {
    return;
  }

  updateNameHelper(currentItem.id, aName, UNDO_STACK);
  currentItem.label = aName;
  currentItem.setAttribute("clipName", aName);
}


function updateNameHelper(aURI, aName, aDestUndoStack)
{
  var subjName = gRDFSvc.GetResource(aURI);
  var predName = gRDFSvc.GetResource("http://clippings.mozdev.org/ns/rdf#name");
  var oldName, targOldName;
  targOldName = gClippings.dataSrc.GetTarget(subjName, predName, true);

  if (targOldName instanceof Components.interfaces.nsIRDFLiteral) {
    oldName = targOldName.Value;
  }

  if (oldName == aName) {
    return;
  }

  if (aDestUndoStack == UNDO_STACK) {
    gUndoStack.push(ACTION_EDITNAME, aURI, oldName, null);
    gClippings.log(String.format("Old Entry name: %S\nOld entry name added to undo stack", oldName));
  }
  else if (aDestUndoStack == REDO_STACK) {
    gRedoStack.push(ACTION_EDITNAME, aURI, oldName, null);
    gClippings.log(String.format("Old Entry name: %S\nOld entry name added to redo stack", oldName));
  }

  var targNewName = gRDFSvc.GetLiteral(aName);
  gClippings.dataSrc.Change(subjName, predName, targOldName, targNewName);
}


function updateText(aText)
{
  var clippingsList = document.getElementById("clippings-list");
  var currentItem = clippingsList.getItemAtIndex(gCurrentListItemIndex);
  if (! currentItem) {
    return;
  }

  updateTextHelper(currentItem.id, aText, UNDO_STACK);
  currentItem.setAttribute("clipText", aText);
}


function updateTextHelper(aURI, aText, aDestUndoStack)
{
  var subjText = gRDFSvc.GetResource(aURI);
  var predText = gRDFSvc.GetResource("http://clippings.mozdev.org/ns/rdf#text");
  var oldText, targOldText;
  targOldText = gClippings.dataSrc.GetTarget(subjText, predText, true);

  if (targOldText instanceof Components.interfaces.nsIRDFLiteral) {
    oldText = targOldText.Value;
  }

  if (oldText == aText) {
    return;
  }

  if (aDestUndoStack == UNDO_STACK) {
    // DEBUGGING
    var subjName = subjText;
    var predName = gRDFSvc.GetResource("http://clippings.mozdev.org/ns/rdf#name");
    var targName = gClippings.dataSrc.GetTarget(subjName, predName, true);
    var name = targName.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
    // END DEBUGGING
    gUndoStack.push(ACTION_EDITTEXT, aURI, name, oldText);
    // We shouldn't care what the old name is.
    // Non-debug invocation to push() should be as follows:
    /***
	gUndoStack.push(ACTION_EDITTEXT, aURI, null, oldText);
    ***/
    gClippings.log(String.format("Entry name: %S\nOld text before edit added to undo stack\noldText = %S", name, oldText));
  }
  else if (aDestUndoStack == REDO_STACK) {
    gRedoStack.push(ACTION_EDITTEXT, aURI, null, oldText);
    gClippings.log(String.format("Added to redo stack: oldText == %S", oldText));
  }

  var targNewText = gRDFSvc.GetLiteral(aText);
  gClippings.dataSrc.Change(subjText, predText, targOldText, targNewText); 
}


function updateCurrentClippingData()
{
  if (gClippings.RDFContainer.GetCount() > 0) {
    updateName(document.getElementById("clipping-name").value);
    updateText(document.getElementById("clipping-text").value);
  }
}


function undo()
{
  updateCurrentClippingData();

  if (gUndoStack.length == 0) {
    gClippings.beep();
    gStatusBar.label = "Nothing to undo";
    return;
  }

  var undone = gUndoStack.pop();

  if (undone.action == ACTION_DELETE) {
    newClippingHelper(undone.name, undone.text, undone.uri);
    selectListItemWithId(undone.uri);
    gClippings.log(String.format("Deletion undone\nEntry name: %S; entry URI: %S", undone.name, undone.uri));
    gRedoStack.push(ACTION_DELETE, undone.uri, undone.name, undone.text);
  }
  else if (undone.action == ACTION_EDITTEXT) {
    updateTextHelper(undone.uri, undone.text, REDO_STACK);

    var listItem = selectListItemWithId(undone.uri);
    if (listItem) {
      updateDisplay();

      var clippingText = document.getElementById("clipping-text");
      clippingText.focus();
      clippingText.select();

      gClippings.log(String.format("Entry text edit undone\nEntry name: %S; entry URI: %S; oldText: %S", undone.name, undone.uri, undone.text));
    }
  }
  else if (undone.action == ACTION_EDITNAME) {
    updateNameHelper(undone.uri, undone.name, REDO_STACK);

    var listItem = selectListItemWithId(undone.uri);
    if (listItem) {
      updateDisplay();
      listItem.label = undone.name;

      var clippingName = document.getElementById("clipping-name");
      clippingName.focus();
      clippingName.select();

      gClippings.log(String.format("Entry name edit undone\nEntry name: %S; entry URI:", undone.name, undone.uri));
    }
  }
  else if (undone.action == ACTION_CREATENEW) {
    var listItem = selectListItemWithId(undone.uri);
    deleteClippingHelper(listItem);
    gClippings.log(String.format("Entry creation undone\nEntry name: %S; entry URI: %S", undone.name, undone.uri));
    gRedoStack.push(ACTION_CREATENEW, undone.uri, undone.name, undone.text);
  }
}


function selectListItemWithId(aID)
{
  var rv;
  var clippingsList = document.getElementById("clippings-list");
  var listItems = clippingsList.getElementsByAttribute("id", aID);

  if (listItems) {
    var listItem = listItems[0];
    clippingsList.selectItem(listItem);
    clippingsList.ensureElementIsVisible(listItem);
    rv = listItem;
  }
  return rv;
}


function reverseLastUndo()
{
  updateCurrentClippingData();

  if (gRedoStack.length == 0) {
    gClippings.beep();
    return;
  }

  var lastUndo = gRedoStack.pop();

  if (lastUndo.action == ACTION_DELETE) {
    var listItem = selectListItemWithId(lastUndo.uri);
    deleteClippingHelper(listItem, UNDO_STACK);
    gClippings.log(String.format("Entry deletion redone!\nEntry name: %S; entry URI: %S", lastUndo.name, lastUndo.uri));
  }
  else if (lastUndo.action == ACTION_EDITNAME) {
    updateNameHelper(lastUndo.uri, lastUndo.name, UNDO_STACK);

    var listItem = selectListItemWithId(lastUndo.uri);
    if (listItem) {
      updateDisplay();
      listItem.label = lastUndo.name;

      var clippingName = document.getElementById("clipping-name");
      clippingName.focus();
      clippingName.select();

      gClippings.log(String.format("Entry name edit redone!\nEntry name: %S; entry URI:", lastUndo.name, lastUndo.uri));
    }
  }
  else if (lastUndo.action == ACTION_EDITTEXT) {
    updateTextHelper(lastUndo.uri, lastUndo.text, UNDO_STACK);

    var listItem = selectListItemWithId(lastUndo.uri);
    if (listItem) {
      updateDisplay();

      var clippingText = document.getElementById("clipping-text");
      clippingText.focus();
      clippingText.select();

      gClippings.log(String.format("Entry text edit redone!\nEntry name: %S; entry URI: %S; oldText: %S", lastUndo.name, lastUndo.uri, lastUndo.text));
    }
  }
  else if (lastUndo.action == ACTION_CREATENEW) {
    newClippingHelper(lastUndo.name, lastUndo.text, lastUndo.uri, UNDO_STACK);
    selectListItemWithId(lastUndo.uri);
    gClippings.log(String.format("Deletion redone!\nEntry name: %S; entry URI: %S", lastUndo.name, lastUndo.uri));
  }

  gStatusBar.label = "Undo Undo!";
}


function showHelp() 
{
  updateCurrentClippingData();
  alert("No Help is available (so leave me alone)");
}
